"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const assert_1 = require("assert");
const _ = require("lodash");
const inspect_1 = require("./inspect");
const utils = require("./utils");
const locations = {};
function isPromise(promise) {
    // eslint-disable-next-line promise/prefer-await-to-then
    return promise && typeof promise.then === "function";
}
function type(value) {
    const natives = {
        "[object Arguments]": "arguments",
        "[object Array]": "array",
        "[object AsyncFunction]": "function",
        "[object Date]": "date",
        "[object Function]": "function",
        "[object Number]": "number",
        "[object RegExp]": "regexp",
        "[object String]": "string"
    };
    if (value === null) {
        return "null";
    }
    if (value === undefined) {
        return "undefined";
    }
    if (Buffer.isBuffer(value)) {
        return "buffer";
    }
    const name = Object.prototype.toString.call(value);
    if (natives[name]) {
        return natives[name];
    }
    if (value === new Object(value)) {
        return "object";
    }
    return typeof value;
}
function display(value) {
    let string = inspect_1.default(value);
    if (isPromise(value))
        string = "[Promise]";
    if (value instanceof Error) {
        string = `[${value.toString()}]`;
    }
    if (string.length <= 40)
        return string;
    if (_.isArray(value)) {
        return `[Array(${value.length})]`;
    }
    if (typeof value === "object") {
        const keys = _.keys(value);
        if (keys.length > 2) {
            return `{ Object (${keys.splice(0, 2).join(", ")}, ...) }`;
        }
        return `{ Object (${keys.join(", ")}) }`;
    }
    return `${string.slice(0, 40)}...'`;
}
function assert(assertion, condition, error) {
    if (!condition) {
        delete locations[assertion.location];
        throw new assert_1.AssertionError({
            "message": error,
            "actual": false,
            "expected": true,
            "operator": "==",
            "stackStartFunction": assert
        });
    }
}
function thrownAt(error = new Error()) {
    let stack = "";
    if (typeof error.stack === "string") {
        //@ts-ignore
        ({ stack } = error);
    }
    const frame = _(stack).replace(error.toString(), "").split("\n").slice(1).filter((line) => !_.includes(line, __dirname))[0] || "";
    const at = frame.match(/^\s*at \(?(.+):(\d+):(\d+)\)?$/);
    if (_.isArray(at)) {
        return {
            "filename": at[1],
            "line": at[2],
            "column": at[3]
        };
    }
    return undefined;
}
exports.thrownAt = thrownAt;
function fail(message) {
    throw new Error(message);
}
exports.fail = fail;
class Assertion {
    constructor(ref, prefix = "", location) {
        this.flags = {};
        this.ref = ref;
        this.prefix = prefix;
        this.location = location;
    }
    assert(result, verb, actual, expected) {
        delete locations[this.location];
        if (this.flags.not && !result) {
            this.flags = {};
            return this;
        }
        if (!this.flags.not && result) {
            this.flags = {};
            return this;
        }
        if (verb === "exist" && this.flags.not && this.ref instanceof Error) {
            const original = this.ref;
            original.at = thrownAt();
            throw original;
        }
        let message = "";
        if (this.prefix) {
            message += `${this.prefix}: `;
        }
        message += `Expected ${display(this.ref)} to `;
        if (this.flags.not) {
            message += "not ";
        }
        message += verb;
        if (this.flags.once) {
            message += " once";
        }
        if (arguments.length === 3) {
            // 'actual' without 'expected'
            message += ` but got ${display(actual)}`;
        }
        class AssertError extends Error {
            constructor(message, actual, expected, assert) {
                super(message);
                this.name = "AssertError";
                Error.captureStackTrace(this, assert);
                this.actual = actual;
                this.expected = expected;
                this.at = thrownAt(this);
            }
        }
        throw new AssertError(message, actual, expected, this.assert);
    }
    NaN() {
        return this.assert(Number.isNaN(this.ref), "be NaN");
    }
    arguments() {
        return this.assertType("arguments");
    }
    array() {
        return this.assertType("array");
    }
    boolean() {
        return this.assertType("boolean");
    }
    buffer() {
        return this.assertType("buffer");
    }
    date() {
        return this.assertType("date");
    }
    function() {
        return this.assertType("function");
    }
    number() {
        return this.assertType("number");
    }
    regexp() {
        return this.assertType("regexp");
    }
    string() {
        return this.assertType("string");
    }
    object() {
        return this.assertType("object");
    }
    assertType(word) {
        let article = "a ";
        if (_.includes(["a", "e", "i", "o", "u"], word[0]))
            article = "an ";
        const typeStr = type(this.ref);
        return this.assert(typeStr === word, `be ${article}${word}`, typeStr);
    }
    /* type, message */
    error(...args) {
        const lastArg = args[1] || args[0];
        const err = this.ref;
        let message = null;
        if (typeof lastArg === "string" || lastArg instanceof RegExp)
            message = lastArg;
        let type = Error;
        if (args.length && typeof args[0] !== "string" && !(args[0] instanceof RegExp))
            ([type] = args);
        if (!this.flags.not || message === null) {
            this.assert(err instanceof type, `be an error with ${type.name || "provided"} type`);
        }
        if (message !== null) {
            const error = err.message || "";
            if (typeof message === "string") {
                this.assert(error === message, "be an error with specified message", error, message);
            }
            else {
                this.assert(error.match(message), "be an error with specified message", error, message);
            }
        }
    }
    true() {
        return this.assert(this.ref === true, "be true");
    }
    false() {
        return this.assert(this.ref === false, "be false");
    }
    null() {
        return this.assert(this.ref === null, "be null");
    }
    undefined() {
        return this.assert(this.ref === undefined, "be undefined");
    }
    include(value) {
        assert(this, arguments.length === 1, "Can only assert include with a single parameter");
        this.flags.deep = !this.flags.shallow;
        if (!{}.hasOwnProperty.call(this.flags, "part"))
            this.flags.part = false;
        return this.assert(utils.contain(this.ref, value, this.flags), `include ${display(value)}`);
    }
    includes(value) {
        return this.include(value);
    }
    contain(value) {
        return this.include(value);
    }
    contains(value) {
        return this.include(value);
    }
    endWith(value) {
        assert(this, typeof this.ref === "string" && typeof value === "string", "Can only assert endsWith on a string, with a string");
        const comparator = this.ref.slice(-value.length);
        return this.assert(comparator === value, `endWith ${display(value)}`);
    }
    endsWith(value) {
        return this.endWith(value);
    }
    startWith(value) {
        assert(this, typeof this.ref === "string" && typeof value === "string", "Can only assert startsWith on a string, with a string");
        const comparator = this.ref.slice(0, value.length);
        return this.assert(comparator === value, `startWith ${display(value)}`);
    }
    startsWith(value) {
        return this.startWith(value);
    }
    exist() {
        return this.assert(!_.isNil(this.ref), "exist");
    }
    exists() {
        return this.exist();
    }
    empty() {
        assert(this, typeof this.ref === "object" || typeof this.ref === "string", "Can only assert empty on object, array or string");
        let { length } = _.keys(this.ref);
        if (this.ref.length !== undefined)
            ({ length } = this.ref);
        return this.assert(!length, "be empty");
    }
    length(size) {
        assert(this, (typeof this.ref === "object" && this.ref !== null) || typeof this.ref === "string", "Can only assert length on object, array or string");
        let { length } = _.keys(this.ref);
        if (this.ref.length !== undefined)
            ({ length } = this.ref);
        return this.assert(length === size, `have a length of ${size}`, length);
    }
    equal(value, options = {}) {
        options.prototype;
        // eslint-disable-next-line func-style
        let compare = (a, b) => utils.deepEqual(a, b, { "prototype": options.prototype });
        if (this.flags.shallow)
            compare = (a, b) => a === b;
        return this.assert(compare(this.ref, value), `equal specified value: ${display(value)}`, this.ref, value);
    }
    equals(value, options = {}) {
        return this.equal(value, options);
    }
    above(value) {
        return this.assert(this.ref > value, `be above ${value}`);
    }
    greaterThan(value) {
        return this.above(value);
    }
    least(value) {
        return this.assert(this.ref >= value, `be at least ${value}`);
    }
    min(value) {
        return this.least(value);
    }
    below(value) {
        return this.assert(this.ref < value, `be below ${value}`);
    }
    lessThan(value) {
        return this.below(value);
    }
    most(value) {
        return this.assert(this.ref <= value, `be at most ${value}`);
    }
    max(value) {
        return this.most(value);
    }
    within(from, to) {
        return this.assert(this.ref >= from && this.ref <= to, `be within ${from}..${to}`);
    }
    range(from, to) {
        return this.within(from, to);
    }
    between(from, to) {
        return this.assert(this.ref > from && this.ref < to, `be between ${from}..${to}`);
    }
    about(value, delta) {
        assert(this, type(this.ref) === "number", "Can only assert about on numbers");
        assert(this, type(value) === "number" && type(delta) === "number", "About assertion requires two number arguments");
        return this.assert(Math.abs(this.ref - value) <= delta, `be about ${value} \u00B1${delta}`);
    }
    instanceof(type) {
        return this.assert(this.ref instanceof type, `be an instance of ${type.name || "provided type"}`);
    }
    instanceOf(type) {
        return this.instanceof(type);
    }
    match(regex) {
        return this.assert(!_.isNull(regex.exec(this.ref)), `match ${regex}`);
    }
    matches(regex) {
        return this.match(regex);
    }
    satisfy(validator) {
        return this.assert(validator(this.ref), "satisfy rule");
    }
    satisfies(validator) {
        return this.satisfy(validator);
    }
    /* type, message */
    throw(...args) {
        assert(this, typeof this.ref === "function", "Can only assert throw on functions");
        assert(this, !this.flags.not || !args.length, "Cannot specify arguments when expecting not to throw");
        const lastArg = args[1] || args[0];
        let type = null;
        if (args.length && typeof args[0] !== "string" && !(args[0] instanceof RegExp))
            ([type] = args);
        let message = null;
        if (typeof lastArg === "string" || lastArg instanceof RegExp)
            message = lastArg;
        let thrown = false;
        try {
            this.ref();
        }
        catch (err) {
            thrown = true;
            if (type) {
                this.assert(err instanceof type, `throw ${type.name || "provided type"}`);
            }
            if (message !== null) {
                const error = err.message || "";
                if (typeof message === "string") {
                    this.assert(error === message, "throw an error with specified message", error, message);
                }
                else {
                    this.assert(error.match(message), "throw an error with specified message", error, message);
                }
            }
            this.assert(thrown, "throw an error", err);
        }
        return this.assert(thrown, "throw an error");
    }
    /* type, message */
    throws(...args) {
        return this.throw(...args);
    }
    /* type, message */
    async reject(...args) {
        try {
            assert(this, isPromise(this.ref), "Can only assert reject on promises");
            const lastArg = args[1] || args[0];
            let type = null;
            if (args.length && typeof args[0] !== "string" && !(args[0] instanceof RegExp))
                ([type] = args);
            let message = null;
            if (typeof lastArg === "string" || lastArg instanceof RegExp)
                message = lastArg;
            let thrown = null;
            try {
                await this.ref;
            }
            catch (err) {
                thrown = err;
            }
            assert(this, !this.flags.not || !args.length, "Cannot specify arguments when expecting not to reject");
            if (thrown) {
                if (type) {
                    this.assert(thrown instanceof type, `reject with ${type.name || "provided type"}`);
                }
                if (message !== null) {
                    const error = thrown.message || "";
                    if (typeof message === "string") {
                        this.assert(error === message, "reject with an error with specified message", error, message);
                    }
                    else {
                        this.assert(!_.isNull(error.match(message)), "reject with an error with specified message", error, message);
                    }
                }
                this.assert(!_.isNull(thrown), "reject with an error", thrown);
            }
            this.assert(!_.isNull(thrown), "reject with an error");
            return thrown;
        }
        catch (err) {
            return new Promise((resolve, reject) => {
                reject(err);
            });
        }
    }
    /* type, message */
    async rejects(...args) {
        return this.reject(...args);
    }
    get not() {
        this.flags["not"] = !this.flags["not"];
        return this;
    }
    get once() {
        this.flags["once"] = !this.flags["once"];
        return this;
    }
    get only() {
        this.flags["only"] = !this.flags["only"];
        return this;
    }
    get part() {
        this.flags["part"] = !this.flags["part"];
        return this;
    }
    get shallow() {
        this.flags["shallow"] = !this.flags["shallow"];
        return this;
    }
    get a() {
        return this;
    }
    get an() {
        return this;
    }
    get and() {
        return this;
    }
    get at() {
        return this;
    }
    get be() {
        return this;
    }
    get have() {
        return this;
    }
    get in() {
        return this;
    }
    get to() {
        return this;
    }
}
exports.Assertion = Assertion;
let c = 0;
function expect(value, prefix) {
    const at = thrownAt();
    const location = `${at.filename}:${at.line}.${at.column}`;
    locations[location] = true;
    ++c;
    return new Assertion(value, prefix, location);
}
exports.expect = expect;
function count() {
    return c;
}
exports.count = count;
exports.default = {
    expect,
    thrownAt,
    fail,
    count
};
//# sourceMappingURL=code.js.map